import json
import re
from argparse import Action, ArgumentParser, Namespace
from typing import Any, Iterable, List, Optional, Sequence, Union

import shtab

from ldb.op_type import OpType

TAG_PATTERN = r"^[^\s,]+$"


def tag_list(value: str) -> List[str]:
    result = []
    for item in value.split(","):
        item = item.strip()
        item_match = re.search(TAG_PATTERN, item)
        if item_match is None:
            raise ValueError(
                f"Invalid tag: {json.dumps(item)}\n"
                "Tags may not contain commas or whitespace characters",
            )
        result.append(item_match.group())
    return result


class AppendConstValuesAction(Action):
    def __call__(
        self,
        parser: ArgumentParser,
        namespace: Namespace,
        values: Union[str, Sequence[Any], None],
        option_string: Optional[str] = None,
    ) -> None:
        items = getattr(namespace, self.dest, None) or []
        items.append((self.const, values))
        setattr(namespace, self.dest, items)


def choice_str(choices: Iterable[str]) -> str:
    choice_strings = ",".join(str(c) for c in choices)
    return f"{{{choice_strings}}}"


def add_data_obj_params(parser: ArgumentParser, dest: str) -> None:
    add_base_data_object_options(parser, dest)
    add_data_object_paths(parser)


def add_base_data_object_options(parser: ArgumentParser, dest: str) -> None:
    parser.add_argument(
        "--query",
        metavar="<query>",
        const=OpType.ANNOTATION_QUERY,
        default=[],
        dest=dest,
        action=AppendConstValuesAction,
        help="JMESPath-like query applied to annotations",
    )
    parser.add_argument(
        "--jquery",
        metavar="<query>",
        const=OpType.JP_ANNOTATION_QUERY,
        default=[],
        dest=dest,
        action=AppendConstValuesAction,
        help="Fully compliant JMESPath query applied to annotations",
    )
    parser.add_argument(
        "--file",
        metavar="<query>",
        const=OpType.FILE_QUERY,
        default=[],
        dest=dest,
        action=AppendConstValuesAction,
        help="JMESPath-like query applied to data object file attributes",
    )
    parser.add_argument(
        "--path",
        metavar="<pattern>",
        const=OpType.PATH_QUERY,
        default=[],
        dest=dest,
        action=AppendConstValuesAction,
        help=(
            "Python regular expression which must match one of the object's "
            "indexing paths"
        ),
    )
    parser.add_argument(
        "--tag",
        metavar="<tag>",
        const=OpType.TAG_QUERY,
        default=[],
        dest=dest,
        type=tag_list,
        action=AppendConstValuesAction,
        help=(
            "Comma-separated list of tags. Select only data objects that "
            "contain at least one."
        ),
    )
    parser.add_argument(
        "--no-tag",
        metavar="<tag>",
        const=OpType.NO_TAG_QUERY,
        default=[],
        dest=dest,
        type=tag_list,
        action=AppendConstValuesAction,
        help=(
            "Comma-separated list of tags. Select only data objects where at "
            "least one of these tags is missing."
        ),
    )
    parser.add_argument(
        "--limit",
        metavar="<n>",
        const=OpType.LIMIT,
        default=[],
        dest=dest,
        action=AppendConstValuesAction,
        type=int,
        help="Take the first n items",
    )
    parser.add_argument(
        "--sample",
        metavar="<probability>",
        const=OpType.SAMPLE,
        default=[],
        dest=dest,
        action=AppendConstValuesAction,
        type=float,
        help=(
            "Each item will have the given probability of being selected. "
            "Must be a float between 0.0 and 1.0 inclusive"
        ),
    )
    parser.add_argument(
        "--pipe",
        nargs="+",
        metavar="<exec>",
        const=OpType.PIPE,
        default=[],
        dest=dest,
        action=AppendConstValuesAction,
        help="Executable to filter or sort data objects",
    )


def add_data_object_paths(parser: ArgumentParser) -> None:
    parser.add_argument(  # type: ignore[attr-defined]
        "paths",
        metavar="<path>",
        nargs="*",
        help="Storage location, data object identifier, or dataset",
    ).complete = shtab.FILE


def add_data_format_arguments(
    parser: ArgumentParser,
    default: str,
    formats: Iterable[str],
) -> None:
    parser.add_argument(
        "-m",
        "--format",
        default=default,
        metavar="<format>",
        choices=formats,
        help=(
            f"Data format to use. (default: {default}) "
            f"Options: {choice_str(formats)}"
        ),
    )


def add_target_dir_argument(parser: ArgumentParser) -> None:
    parser.add_argument(  # type: ignore[attr-defined]
        "-t",
        "--target",
        dest="target_dir",
        default=".",
        metavar="<dir>",
        help="Target directory",
    ).complete = shtab.DIR



import importlib
import os
from types import ModuleType


def import_path(path: str, name: str = "") -> ModuleType:
    """
    Import a python file as a single-use module.

    This is intended for single-use imports and does not add the module to
    `sys.modules`. For files that may be imported more than once, make sure
    they're on the python path (`sys.path`) and use a regular import statement.

    See "Importing a source file directly" in the `importlib` documentation:
    https://docs.python.org/3.7/library/importlib.html#importing-a-source-file-directly
    """
    if not name:
        name = os.path.splitext(os.path.basename(path))[0]
    spec = importlib.util.spec_from_file_location(name, path)
    if spec is None or spec.loader is None:
        raise ImportError("Could not load module with args: {name=}, {path=}")
    module = importlib.util.module_from_spec(spec)
    loader: importlib.abc.Loader = spec.loader
    loader.exec_module(module)
    return module



from argparse import ArgumentParser, Namespace
from typing import TYPE_CHECKING, Iterable

from ldb.cli_utils import add_data_obj_params
from ldb.core import get_ldb_instance
from ldb.ls import ls, print_dataset_listings

if TYPE_CHECKING:
    from argparse import _SubParsersAction


def ls_command(options: Namespace) -> None:
    ldb_dir = get_ldb_instance()
    ds_listings = ls(
        ldb_dir,
        options.paths,
        options.query_args,
        warn=False,
    )
    if options.summary:
        print(len(ds_listings))
    else:
        num_items = print_dataset_listings(
            ds_listings,
            verbose=options.verbose,
        )
        print(f"\n{num_items} matching data objects")


def add_parser(
    subparsers: "_SubParsersAction[ArgumentParser]",
    parents: Iterable[ArgumentParser],
) -> None:
    parser = subparsers.add_parser(
        "list",
        parents=parents,
        help="List objects and annotations",
    )
    parser.add_argument(
        "-s",
        "--summary",
        action="store_true",
        default=False,
        help="Show the number of objects instead of listing them out",
    )
    add_data_obj_params(parser, dest="query_args")
    parser.set_defaults(func=ls_command)



import json
import sys

from .resnet_utils import image_layer_similarity
from .utils import sort_by_iterable


def main() -> None:
    if len(sys.argv) < 2:
        print(
            "usage: resnet-image <image_file_path> [<model_number>]",
            file=sys.stderr,
        )
        sys.exit(1)

    image_filepath = sys.argv[1]
    model_name = "18"
    layer_index = None
    if len(sys.argv) > 2:
        model_name = sys.argv[2]
        if len(sys.argv) > 3:
            layer_index = int(sys.argv[3])
            if not 1 <= layer_index <= 4:
                raise ValueError(
                    "Layer number must be in the range (1, 4) inclusive",
                )
            layer_index -= 1
    data = json.loads(sys.stdin.read())
    file_paths = [d[1] for d in data]
    similarity = image_layer_similarity(
        image_filepath,
        file_paths,
        model_name,
        layer_index,
    )
    for data_object_hash, _, _ in sort_by_iterable(data, similarity):
        print(data_object_hash, flush=True)


if __name__ == "__main__":
    main()



from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import TYPE_CHECKING, Iterable

from ldb.commit import commit
from ldb.core import get_ldb_instance

if TYPE_CHECKING:
    from argparse import _SubParsersAction


def commit_command(options: Namespace) -> None:
    commit(get_ldb_instance(), Path("."), options.message)


def add_parser(
    subparsers: "_SubParsersAction[ArgumentParser]",
    parents: Iterable[ArgumentParser],
) -> None:
    parser = subparsers.add_parser(
        "commit",
        parents=parents,
        help="Commit the currently staged workspace dataset",
    )
    parser.add_argument(
        "message",
        metavar="<message>",
        nargs="?",
        default="",
        help="A message about this commit",
    )
    parser.set_defaults(func=commit_command)



from typing import Any

from jmespath import exceptions


class IdentifierException(exceptions.JMESPathError):
    def __init__(  # pylint: disable=super-init-not-called
        self,
        identifier: str,
    ) -> None:
        self.identifier = identifier

    def __str__(self) -> str:
        return f"{self.identifier}"


class MissingIdentifierException(IdentifierException):
    pass


class IdentifierTypeException(IdentifierException):
    def __init__(self, identifier: str, subject: Any) -> None:
        super().__init__(identifier)
        self.subject = subject

    def __str__(self) -> str:
        return (
            f"Invalid application of identifier {self.identifier}. "
            "Identifier expression subject must have a JSON Object-like type "
            "such as dict, not "
            f"{type(self.subject)}"
        )


class InvalidIdentifierExpression(IdentifierException):
    pass



from dataclasses import dataclass
from enum import IntEnum
from pathlib import Path
from typing import Dict, Iterable, Iterator, Optional, Tuple

from ldb.dataset import Dataset, get_collection
from ldb.exceptions import LDBException
from ldb.path import InstanceDir, WorkspacePath
from ldb.utils import (
    DATASET_PREFIX,
    WORKSPACE_DATASET_PREFIX,
    format_dataset_identifier,
    get_hash_path,
    load_data_file,
    parse_dataset_identifier,
)
from ldb.workspace import collection_dir_to_object, load_workspace_dataset


class DiffType(IntEnum):
    SAME = 1
    ADDITION = 2
    DELETION = 3
    MODIFICATION = 4


@dataclass
class SimpleDiffItem:
    data_object_hash: str
    annotation_hash1: str
    annotation_hash2: str
    diff_type: DiffType


@dataclass
class DiffItem(SimpleDiffItem):
    data_object_path: str
    annotation_version1: int
    annotation_version2: int


def get_diff_collection(
    ldb_dir: Path,
    dataset: str,
) -> Dict[str, Optional[str]]:
    if dataset.startswith(DATASET_PREFIX):
        return get_collection(
            ldb_dir,
            get_dataset_version_hash(ldb_dir, dataset),
        )
    if dataset.startswith(WORKSPACE_DATASET_PREFIX):
        return collection_dir_to_object(
            Path(dataset[len(WORKSPACE_DATASET_PREFIX) :])  # noqa: E203
            / WorkspacePath.COLLECTION,
        )
    return get_collection(ldb_dir, dataset)


def get_diff_collections(
    ldb_dir: Path,
    dataset1: str = "",
    dataset2: str = "",
    workspace_path: str = ".",
) -> Tuple[Dict[str, Optional[str]], Dict[str, Optional[str]]]:
    if dataset1:
        collection1 = get_diff_collection(ldb_dir, dataset1)
        if dataset2:
            return collection1, get_diff_collection(ldb_dir, dataset2)
    elif dataset2:
        raise Exception
    else:
        workspace_dataset = load_workspace_dataset(Path(workspace_path))
        if workspace_dataset.parent:
            collection1 = get_collection(ldb_dir, workspace_dataset.parent)
        else:
            collection1 = {}
    collection2 = collection_dir_to_object(
        Path(workspace_path) / WorkspacePath.COLLECTION,
    )
    return collection1, collection2


def diff(
    ldb_dir: Path,
    dataset1: str = "",
    dataset2: str = "",
    workspace_path: str = ".",
) -> Iterator[DiffItem]:
    return full_diff(
        ldb_dir,
        simple_diff(
            ldb_dir,
            dataset1,
            dataset2,
            workspace_path,
        ),
    )


def summarize_diff(
    diff_items: Iterable[SimpleDiffItem],
) -> Tuple[int, int, int]:
    additions = 0
    deletions = 0
    modifications = 0
    for item in diff_items:
        if item.diff_type == DiffType.ADDITION:
            additions += 1
        elif item.diff_type == DiffType.DELETION:
            deletions += 1
        elif item.diff_type == DiffType.MODIFICATION:
            modifications += 1
    return additions, deletions, modifications


def get_dataset_version_hash(ldb_dir: Path, dataset_identifier: str) -> str:
    ds_name, ds_version_num = parse_dataset_identifier(dataset_identifier)
    dataset_obj = Dataset.parse(
        load_data_file(ldb_dir / InstanceDir.DATASETS / ds_name),
    )
    if ds_version_num is None:
        return dataset_obj.versions[-1]
    try:
        return dataset_obj.versions[ds_version_num - 1]
    except IndexError as exc:
        latest_dataset = format_dataset_identifier(
            ds_name,
            len(dataset_obj.versions),
        )
        raise LDBException(
            f"{dataset_identifier} does not exist\n"
            f"The latest version is {latest_dataset}",
        ) from exc


def full_diff(
    ldb_dir: Path,
    simple_diff_items: Iterable[SimpleDiffItem],
) -> Iterator[DiffItem]:
    data_object_info_path = ldb_dir / InstanceDir.DATA_OBJECT_INFO
    for item in simple_diff_items:
        annotation_version1 = get_annotation_version(
            ldb_dir,
            item.data_object_hash,
            item.annotation_hash1,
        )
        if item.diff_type == DiffType.SAME:
            annotation_version2 = annotation_version1
        else:
            annotation_version2 = get_annotation_version(
                ldb_dir,
                item.data_object_hash,
                item.annotation_hash2,
            )
        data_object_meta = load_data_file(
            get_hash_path(
                data_object_info_path,
                item.data_object_hash,
            )
            / "meta",
        )
        yield DiffItem(
            data_object_hash=item.data_object_hash,
            annotation_hash1=item.annotation_hash1,
            annotation_hash2=item.annotation_hash2,
            diff_type=item.diff_type,
            data_object_path=data_object_meta["fs"]["path"],
            annotation_version1=annotation_version1,
            annotation_version2=annotation_version2,
        )


def get_annotation_version(
    ldb_dir: Path,
    data_object_hash: str,
    annotation_hash: str,
) -> int:
    if annotation_hash:
        annotation_meta = load_data_file(
            get_hash_path(
                ldb_dir / InstanceDir.DATA_OBJECT_INFO,
                data_object_hash,
            )
            / "annotations"
            / annotation_hash,
        )
        return annotation_meta["version"]  # type: ignore[no-any-return]
    return 0


def simple_diff(
    ldb_dir: Path,
    dataset1: str = "",
    dataset2: str = "",
    workspace_path: str = ".",
) -> Iterator[SimpleDiffItem]:
    collection1, collection2 = get_diff_collections(
        ldb_dir,
        dataset1,
        dataset2,
        workspace_path,
    )
    return simple_diff_on_collections(collection1, collection2)


def simple_diff_on_collections(
    collection1: Dict[str, Optional[str]],
    collection2: Dict[str, Optional[str]],
) -> Iterator[SimpleDiffItem]:
    # pylint: disable=invalid-name
    iter1 = iter(sorted(collection1.items()))
    iter2 = iter(sorted(collection2.items()))

    d1, a1 = next(iter1, (None, None))
    d2, a2 = next(iter2, (None, None))
    while d1 is not None and d2 is not None:
        if d1 < d2:
            yield SimpleDiffItem(d1, a1 or "", "", DiffType.DELETION)
            d1, a1 = next(iter1, (None, None))
        elif d1 > d2:
            yield SimpleDiffItem(d2, "", a2 or "", DiffType.ADDITION)
            d2, a2 = next(iter2, (None, None))
        else:
            diff_type = DiffType.SAME if a1 == a2 else DiffType.MODIFICATION
            yield SimpleDiffItem(d1, a1 or "", a2 or "", diff_type)
            d1, a1 = next(iter1, (None, None))
            d2, a2 = next(iter2, (None, None))

    if d1 is None:
        if d2 is not None:
            yield SimpleDiffItem(d2, "", a2 or "", DiffType.ADDITION)
    elif d2 is None:
        yield SimpleDiffItem(d1, a1 or "", "", DiffType.DELETION)

    for d1, a1 in iter1:
        yield SimpleDiffItem(d1, a1 or "", "", DiffType.DELETION)
    for d2, a2 in iter2:
        yield SimpleDiffItem(d2, "", a2 or "", DiffType.ADDITION)


def format_summary(additions: int, deletions: int, modifications: int) -> str:
    return (
        f"  Additions (+): {additions:8}\n"
        f"  Deletions (-): {deletions:8}\n"
        f"  Modifications (m): {modifications:4}"
    )



from pathlib import Path
from typing import Collection, Sequence

from ldb.data_formats import INDEX_FORMATS, Format
from ldb.index.annotation_only import AnnotationOnlyIndexer
from ldb.index.base import Indexer, IndexingResult, PairIndexer, Preprocessor
from ldb.index.inferred import InferredIndexer, InferredPreprocessor
from ldb.index.label_studio import LabelStudioIndexer, LabelStudioPreprocessor
from ldb.index.utils import FSPathsMapping
from ldb.storage import get_storage_locations


def index(
    ldb_dir: Path,
    paths: Sequence[str],
    read_any_cloud_location: bool = False,
    fmt: str = Format.AUTO,
    tags: Collection[str] = (),
) -> IndexingResult:
    indexer: Indexer
    fmt = INDEX_FORMATS.get(fmt, fmt)
    print(f"Data format: {fmt}")
    storage_locations = get_storage_locations(ldb_dir)
    if fmt in (Format.AUTO, Format.STRICT, Format.BARE, Format.ANNOT):
        preprocessor = Preprocessor(paths, storage_locations)
        if fmt == Format.AUTO:
            fmt = autodetect_format(
                preprocessor.data_object_paths,
                preprocessor.annotation_paths,
            )
            print(f"Auto-detected data format: {fmt}")
        if fmt in (Format.STRICT, Format.BARE):
            indexer = PairIndexer(
                ldb_dir,
                preprocessor,
                read_any_cloud_location,
                fmt == Format.STRICT,
                tags=tags,
            )
        elif fmt == Format.ANNOT:
            indexer = AnnotationOnlyIndexer(
                ldb_dir,
                preprocessor,
                tags=tags,
            )
    elif fmt == Format.INFER:
        preprocessor = InferredPreprocessor(paths, storage_locations)
        indexer = InferredIndexer(
            ldb_dir,
            preprocessor,
            read_any_cloud_location,
            fmt == Format.STRICT,
            tags=tags,
        )
    elif fmt == Format.LABEL_STUDIO:
        preprocessor = LabelStudioPreprocessor(paths, storage_locations)
        indexer = LabelStudioIndexer(
            ldb_dir,
            preprocessor,
            read_any_cloud_location,
            fmt == Format.STRICT,
            tags=tags,
        )
    else:
        raise ValueError(f"Not a valid indexing format: {fmt}")
    print("\nIndexing paths...")
    indexer.index()
    return indexer.result


def autodetect_format(
    data_object_paths: FSPathsMapping,
    annotation_paths: FSPathsMapping,
) -> str:
    if any(annotation_paths.values()) and not any(data_object_paths.values()):
        return Format.ANNOT
    return Format.STRICT



from abc import ABC
from pathlib import Path
from typing import Dict, Iterator, Mapping, TypeVar

KT = TypeVar("KT")
VT = TypeVar("VT")


class MappingCache(ABC, Mapping[KT, VT]):
    def __init__(self) -> None:
        self._cache: Dict[KT, VT] = {}

    def __getitem__(self, key: KT) -> VT:
        try:
            return self._cache[key]
        except KeyError:
            value = self.get_new(key)
            self._cache[key] = value
            return value

    def get_new(self, key: KT) -> VT:
        raise NotImplementedError

    def __iter__(self) -> Iterator[KT]:
        return iter(self._cache)

    def __len__(self) -> int:
        return len(self._cache)


class LDBMappingCache(MappingCache[KT, VT]):
    def __init__(self, ldb_dir: Path) -> None:
        super().__init__()
        self.ldb_dir = ldb_dir



from argparse import ArgumentParser, Namespace
from typing import TYPE_CHECKING, Iterable

from ldb.core import get_ldb_instance
from ldb.ds import ds, print_ds_listings

if TYPE_CHECKING:
    from argparse import _SubParsersAction


def ds_command(options: Namespace) -> None:  # pylint: disable=unused-argument
    ldb_dir = get_ldb_instance()
    ds_listings = ds(
        ldb_dir,
    )
    print_ds_listings(ds_listings)


def add_parser(
    subparsers: "_SubParsersAction[ArgumentParser]",
    parents: Iterable[ArgumentParser],
) -> None:
    parser = subparsers.add_parser(
        "ds",
        parents=parents,
        help="List datasets",
    )
    parser.set_defaults(func=ds_command)



from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import TYPE_CHECKING, Iterable

import shtab

from ldb.config import set_default_instance
from ldb.core import init

if TYPE_CHECKING:
    from argparse import _SubParsersAction


def init_command(options: Namespace) -> None:
    init(options.path, force=options.force, read_any_cloud_location=False)
    set_default_instance(options.path, overwrite_existing=False)


def add_parser(
    subparsers: "_SubParsersAction[ArgumentParser]",
    parents: Iterable[ArgumentParser],
) -> None:
    parser = subparsers.add_parser(
        "init",
        parents=parents,
        help="Initialize an empty LDB instance",
    )
    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        default=False,
        help="Overwrite an existing instance",
    )
    parser.add_argument(  # type: ignore[attr-defined]
        "path",
        metavar="<path>",
        type=Path,
        help="Directory in which to initialize new instance",
    ).complete = shtab.DIRECTORY
    parser.set_defaults(func=init_command)



import os
import re
from enum import Enum, unique
from itertools import tee
from pathlib import Path
from typing import (
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    NamedTuple,
    Sequence,
    Set,
    Tuple,
)

from fsspec.core import get_fs_token_paths
from fsspec.spec import AbstractFileSystem
from fsspec.utils import get_protocol
from tomlkit import document
from tomlkit.toml_document import TOMLDocument

from ldb import config
from ldb.config import ConfigType
from ldb.core import get_ldb_instance
from ldb.dataset import (
    OpDef,
    apply_queries,
    combine_collections,
    get_collection_dir_items,
    get_collection_dir_keys,
    get_collection_from_dataset_identifier,
    iter_collection_dir,
)
from ldb.exceptions import DataObjectNotFoundError, LDBException
from ldb.index import index
from ldb.index.utils import expand_indexing_paths
from ldb.path import InstanceDir, WorkspacePath
from ldb.storage import StorageLocation, get_storage_locations
from ldb.utils import (
    DATA_OBJ_ID_PATTERN,
    DATA_OBJ_ID_PREFIX,
    DATASET_PREFIX,
    ROOT,
    WORKSPACE_DATASET_PREFIX,
    format_dataset_identifier,
    get_file_hash,
    get_hash_path,
    parse_data_object_hash_identifier,
    parse_dataset_identifier,
)
from ldb.workspace import collection_dir_to_object, load_workspace_dataset


@unique
class ArgType(Enum):
    ROOT_DATASET = "root dataset"
    DATASET = "dataset"
    WORKSPACE_DATASET = "workspace_dataset"
    DATA_OBJECT = "data object"
    PATH = "path"


class AddInput(NamedTuple):
    data_object_hashes: Iterable[str]
    annotation_hashes: Iterable[str]
    message: str


def get_arg_type(paths: Sequence[str]) -> ArgType:
    if any(p == f"{DATASET_PREFIX}{ROOT}" for p in paths):
        return ArgType.ROOT_DATASET
    if any(p.startswith(DATASET_PREFIX) for p in paths):
        return ArgType.DATASET
    if any(p.startswith(WORKSPACE_DATASET_PREFIX) for p in paths):
        return ArgType.WORKSPACE_DATASET
    if any(re.search(DATA_OBJ_ID_PATTERN, p) for p in paths):
        return ArgType.DATA_OBJECT
    return ArgType.PATH


def expands_to_workspace(urlpath: str) -> bool:
    if get_protocol(urlpath) != "file":
        return False
    fs, _, paths = get_fs_token_paths(urlpath)
    for path in paths:
        if (
            fs.isdir(fs.sep.join([path, ".ldb_workspace"]))
            and os.path.abspath(path) != os.getcwd()
        ):
            return True
    return False


def process_args_for_add(
    ldb_dir: Path,
    paths: Sequence[str],
) -> AddInput:
    if not paths:
        raise LDBException("Must supply path")
    return ADD_FUNCTIONS[get_arg_type(paths)](ldb_dir, paths)


def root_dataset_for_add(
    ldb_dir: Path,
    paths: Sequence[str],  # pylint: disable=unused-argument
) -> AddInput:
    data_object_hashes = []
    annotation_hashes = []
    for data_object_hash, annotation_hash in get_collection_dir_items(
        ldb_dir / InstanceDir.DATA_OBJECT_INFO,
        is_workspace=False,
    ):
        data_object_hashes.append(data_object_hash)
        annotation_hashes.append(annotation_hash or "")
    return AddInput(data_object_hashes, annotation_hashes, "")


def dataset_for_add(ldb_dir: Path, paths: Sequence[str]) -> AddInput:
    try:
        dataset_identifiers = [parse_dataset_identifier(p) for p in paths]
    except LDBException as exc:
        raise LDBException(
            "All paths must be the same type. "
            "Found path starting with 'ds', but unable "
            "parse all paths as a dataset identifier",
        ) from exc
    collections = [
        get_collection_from_dataset_identifier(
            ldb_dir,
            ds_name,
            ds_version,
        )
        for ds_name, ds_version in dataset_identifiers
    ]
    combined_collection = combine_collections(ldb_dir, collections)
    return AddInput(
        combined_collection.keys(),
        combined_collection.values(),
        "",
    )


def workspace_dataset_for_add(ldb_dir: Path, paths: Sequence[str]) -> AddInput:
    paths = [re.sub(r"^ws:", "", p) for p in paths]
    collections = [
        collection_dir_to_object(
            Path(path) / WorkspacePath.COLLECTION,
        )
        for path in paths
    ]
    combined_collection = combine_collections(ldb_dir, collections)
    return AddInput(
        combined_collection.keys(),
        combined_collection.values(),
        "",
    )


def data_object_for_add(
    ldb_dir: Path,  # pylint: disable=unused-argument
    paths: Sequence[str],
) -> AddInput:
    try:
        data_object_hashes = sorted(
            parse_data_object_hash_identifier(p) for p in paths
        )
    except ValueError as exc:
        raise LDBException(
            "All paths must be the same type. "
            f"Found path starting with '{DATA_OBJ_ID_PREFIX}', but unable "
            "parse all paths as a data object identifier",
        ) from exc
    return AddInput(
        data_object_hashes,
        get_current_annotation_hashes(ldb_dir, data_object_hashes),
        "",
    )


def path_for_add(ldb_dir: Path, paths: Sequence[str]) -> AddInput:
    data_object_hash_iter, data_object_hash_iter2 = tee(
        data_object_hashes_from_path(paths, get_storage_locations(ldb_dir)),
    )
    try:
        annotation_hashes = get_current_annotation_hashes(
            ldb_dir,
            data_object_hash_iter,
        )
    except DataObjectNotFoundError as exc:
        cfg: TOMLDocument = (
            config.load_first([ConfigType.INSTANCE]) or document()
        )
        auto_index: bool = cfg.get("core", {}).get("auto_index", False)
        if not auto_index:
            raise DataObjectNotFoundError(
                f"{exc!s}\n"
                f"Will not index new data: auto_index = {auto_index}",
            ) from exc

        indexing_result = index(
            ldb_dir,
            paths,
            read_any_cloud_location=(
                cfg.get("core", {}).get("read_any_cloud_location", False)
            ),
        )
        data_object_hashes = indexing_result.data_object_hashes
        annotation_hashes = get_current_annotation_hashes(
            ldb_dir,
            data_object_hashes,
        )
        message = indexing_result.summary()
    else:
        data_object_hashes = list(data_object_hash_iter2)
        message = ""

    return AddInput(
        data_object_hashes,
        annotation_hashes,
        message,
    )


ADD_FUNCTIONS: Dict[ArgType, Callable[[Path, Sequence[str]], AddInput]] = {
    ArgType.ROOT_DATASET: root_dataset_for_add,
    ArgType.DATASET: dataset_for_add,
    ArgType.WORKSPACE_DATASET: workspace_dataset_for_add,
    ArgType.DATA_OBJECT: data_object_for_add,
    ArgType.PATH: path_for_add,
}


def add(
    workspace_path: Path,
    paths: Sequence[str],
    query_args: Iterable[OpDef],
) -> None:
    if not paths:
        if not query_args:
            raise LDBException(
                "Must provide either a query or at least one path",
            )
        paths = [f"{DATASET_PREFIX}{ROOT}"]
    ldb_dir = get_ldb_instance()

    workspace_path = Path(os.path.normpath(workspace_path))
    ds_name = load_workspace_dataset(workspace_path).dataset_name
    collection_dir_path = workspace_path / WorkspacePath.COLLECTION
    collection_dir_path.mkdir(exist_ok=True)

    data_object_hashes, annotation_hashes, message = process_args_for_add(
        ldb_dir,
        paths,
    )
    if message:
        print(message)
        print()

    collection = apply_queries(
        ldb_dir,
        data_object_hashes,
        annotation_hashes,
        query_args,
        warn=False,
    )
    print("Adding to working dataset...")

    num_data_objects = add_to_collection_dir(
        collection_dir_path,
        collection,
    )

    ds_ident = format_dataset_identifier(ds_name)
    print(f"Added {num_data_objects} data objects to {ds_ident}")


def add_to_collection_dir(
    collection_dir_path: Path,
    collection: Iterable[Tuple[str, str]],
) -> int:
    to_write = []
    for data_object_hash, annotation_hash in collection:
        to_write.append(
            (
                get_hash_path(collection_dir_path, data_object_hash),
                annotation_hash,
            ),
        )

    num_data_objects = 0
    for collection_member_path, annotation_hash in to_write:
        if collection_member_path.exists():
            with collection_member_path.open("r+") as file:
                existing_annotation_hash = file.read()
                if annotation_hash != existing_annotation_hash:
                    file.seek(0)
                    file.write(annotation_hash)
                    file.truncate()
                    num_data_objects += 1
        else:
            collection_member_path.parent.mkdir(exist_ok=True)
            with collection_member_path.open("w") as file:
                file.write(annotation_hash)
                num_data_objects += 1
    return num_data_objects


def process_args_for_delete(
    ldb_dir: Path,
    paths: Sequence[str],
) -> List[str]:
    if not paths:
        paths = ["."]
        arg_type = ArgType.WORKSPACE_DATASET
    else:
        arg_type = get_arg_type(paths)
    return DELETE_FUNCTIONS[arg_type](ldb_dir, paths)


def root_dataset_for_delete(
    ldb_dir: Path,
    paths: Sequence[str],  # pylint: disable=unused-argument
) -> List[str]:
    return sorted(
        get_collection_dir_keys(
            ldb_dir / InstanceDir.DATA_OBJECT_INFO,
        ),
    )


def dataset_for_delete(ldb_dir: Path, paths: Sequence[str]) -> List[str]:
    try:
        dataset_identifiers = [parse_dataset_identifier(p) for p in paths]
    except LDBException as exc:
        raise LDBException(
            "All paths must be the same type. "
            "Found path starting with 'ds', but unable "
            "parse all paths as a dataset identifier",
        ) from exc
    collections = [
        get_collection_from_dataset_identifier(
            ldb_dir,
            ds_name,
            ds_version,
        )
        for ds_name, ds_version in dataset_identifiers
    ]
    data_objects: Set[str] = set()
    for collection in collections:
        data_objects.update(collection.keys())
    return list(data_objects)


def workspace_dataset_for_delete(
    ldb_dir: Path,
    paths: Sequence[str],
) -> List[str]:
    return list(workspace_dataset_for_add(ldb_dir, paths)[0])


def data_object_for_delete(
    ldb_dir: Path,  # pylint: disable=unused-argument
    paths: Sequence[str],
) -> List[str]:
    try:
        return [parse_data_object_hash_identifier(p) for p in paths]
    except ValueError as exc:
        raise LDBException(
            "All paths must be the same type. "
            f"Found path starting with '{DATA_OBJ_ID_PREFIX}', but unable "
            "parse all paths as a data object identifier",
        ) from exc


def path_for_delete(
    ldb_dir: Path,
    paths: Sequence[str],
) -> List[str]:
    return list(
        data_object_hashes_from_path(paths, get_storage_locations(ldb_dir)),
    )


def get_data_object_storage_files(
    paths: Iterable[str],
    storage_locations: Iterable[StorageLocation],
) -> Iterator[Tuple[AbstractFileSystem, str]]:
    for fs, fs_paths in expand_indexing_paths(
        paths,
        storage_locations,
        default_format=False,
    ).items():
        for path in fs_paths:
            if not path.endswith(".json"):
                yield fs, path


def data_object_hashes_from_path(
    paths: Iterable[str],
    storage_locations: Iterable[StorageLocation],
) -> Iterator[str]:
    for fs, path in get_data_object_storage_files(paths, storage_locations):
        yield get_file_hash(fs, path)


DELETE_FUNCTIONS: Dict[ArgType, Callable[[Path, Sequence[str]], List[str]]] = {
    ArgType.ROOT_DATASET: root_dataset_for_delete,
    ArgType.DATASET: dataset_for_delete,
    ArgType.WORKSPACE_DATASET: workspace_dataset_for_delete,
    ArgType.DATA_OBJECT: data_object_for_delete,
    ArgType.PATH: path_for_delete,
}


def delete(
    workspace_path: Path,
    paths: Sequence[str],
    query_args: Iterable[OpDef],
) -> None:
    if not paths and not query_args:
        raise LDBException(
            "Must provide either a query or at least one path",
        )

    workspace_path = Path(os.path.normpath(workspace_path))
    ds_name = load_workspace_dataset(workspace_path).dataset_name
    ds_ident = format_dataset_identifier(ds_name)
    collection_dir_path = workspace_path / WorkspacePath.COLLECTION

    data_object_hashes = select_data_object_hashes(
        get_ldb_instance(),
        paths,
        query_args,
        warn=False,
    )
    num_deleted = delete_from_collection_dir(
        collection_dir_path,
        data_object_hashes,
    )
    print(f"Deleted {num_deleted} data objects from {ds_ident}")


def delete_from_collection_dir(
    collection_dir_path: Path,
    data_object_hashes: Iterable[str],
) -> int:
    if not collection_dir_path.exists():
        return 0

    num_deleted = 0
    for data_object_hash in data_object_hashes:
        collection_member_path = get_hash_path(
            collection_dir_path,
            data_object_hash,
        )
        if collection_member_path.exists():
            collection_member_path.unlink()
            try:
                collection_member_path.parent.rmdir()
            except OSError:
                pass
            num_deleted += 1
    return num_deleted


def delete_missing_from_collection_dir(
    collection_dir_path: Path,
    data_object_hashes: Iterable[str],
) -> int:
    data_object_hash_set = set(data_object_hashes)
    num_deleted = 0
    for path in iter_collection_dir(collection_dir_path):
        parent, name = os.path.split(path)
        data_object_hash = os.path.basename(parent) + name
        if data_object_hash not in data_object_hash_set:
            collection_member_path = Path(path)  # TODO: don't use Path obj
            collection_member_path.unlink()
            try:
                collection_member_path.parent.rmdir()
            except OSError:
                pass
            num_deleted += 1
    return num_deleted


def get_current_annotation_hash(ldb_dir: Path, data_object_hash: str) -> str:
    data_object_dir = get_hash_path(
        ldb_dir / InstanceDir.DATA_OBJECT_INFO,
        data_object_hash,
    )
    if not data_object_dir.is_dir():
        raise DataObjectNotFoundError(
            f"Data object not found: {DATA_OBJ_ID_PREFIX}{data_object_hash}",
        )
    try:
        return (data_object_dir / "current").read_text()
    except FileNotFoundError:
        return ""


def get_current_annotation_hashes(
    ldb_dir: Path,
    data_object_hashes: Iterable[str],
) -> List[str]:
    return [
        get_current_annotation_hash(ldb_dir, d) for d in data_object_hashes
    ]


def process_args_for_ls(
    ldb_dir: Path,
    paths: Sequence[str],
) -> AddInput:
    if not paths:
        paths = ["."]
        arg_type = ArgType.WORKSPACE_DATASET
    else:
        arg_type = get_arg_type(paths)
    return LS_FUNCTIONS[arg_type](ldb_dir, paths)


def path_for_ls(ldb_dir: Path, paths: Sequence[str]) -> AddInput:
    hashes = []
    for fs, path in get_data_object_storage_files(
        paths,
        get_storage_locations(ldb_dir),
    ):
        data_object_hash = get_file_hash(fs, path)
        try:
            annotation_hash = get_current_annotation_hash(
                ldb_dir,
                data_object_hash,
            )
        except DataObjectNotFoundError as exc:
            raise DataObjectNotFoundError(
                "Data object not found: "
                f"{DATA_OBJ_ID_PREFIX}{data_object_hash} "
                f"(path={path!r})",
            ) from exc
        hashes.append((data_object_hash, annotation_hash))
    if hashes:
        data_object_hashes, annotation_hashes = (
            list(x) for x in zip(*sorted(hashes))
        )
    else:
        data_object_hashes, annotation_hashes = [], []
    return AddInput(
        data_object_hashes,
        annotation_hashes,
        "",
    )


LS_FUNCTIONS: Dict[ArgType, Callable[[Path, Sequence[str]], AddInput]] = {
    ArgType.ROOT_DATASET: root_dataset_for_add,
    ArgType.DATASET: dataset_for_add,
    ArgType.WORKSPACE_DATASET: workspace_dataset_for_add,
    ArgType.DATA_OBJECT: data_object_for_add,
    ArgType.PATH: path_for_ls,
}


def select_data_object_hashes(
    ldb_dir: Path,
    paths: Sequence[str],
    query_args: Iterable[OpDef],
    warn: bool = True,
) -> Iterable[str]:
    if not query_args:
        data_object_hashes: Iterable[str] = process_args_for_delete(
            ldb_dir,
            paths,
        )
    else:
        data_object_hashes, annotation_hashes, _ = process_args_for_ls(
            ldb_dir,
            paths,
        )
        collection = apply_queries(
            ldb_dir,
            data_object_hashes,
            annotation_hashes,
            query_args,
            warn=warn,
        )
        data_object_hashes = (d for d, _ in collection)
    return data_object_hashes


def sync(
    workspace_path: Path,
    paths: Sequence[str],
    query_args: Iterable[OpDef],
) -> None:
    if not paths:
        paths = ["."]

    ldb_dir = get_ldb_instance()
    workspace_path = Path(os.path.normpath(workspace_path))
    ds_name = load_workspace_dataset(workspace_path).dataset_name
    ds_ident = format_dataset_identifier(ds_name)
    collection_dir_path = workspace_path / WorkspacePath.COLLECTION
    collection_dir_path.mkdir(exist_ok=True)

    data_object_hashes, annotation_hashes, message = process_args_for_add(
        ldb_dir,
        paths,
    )
    if message:
        print(message)
        print()

    collection = list(
        apply_queries(
            ldb_dir,
            data_object_hashes,
            annotation_hashes,
            query_args,
            warn=False,
        ),
    )

    data_object_hashes = {d for d, _ in collection}
    print("Syncing working dataset...")
    num_data_objects = add_to_collection_dir(
        collection_dir_path,
        collection,
    )
    print(f"Added {num_data_objects} data objects to {ds_ident}")
    num_deleted = delete_missing_from_collection_dir(
        collection_dir_path,
        data_object_hashes,
    )
    print(f"Deleted {num_deleted} data objects from {ds_ident}")



import os
from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import TYPE_CHECKING, Iterable

from ldb.core import get_ldb_instance
from ldb.dataset import get_collection_size
from ldb.diff import format_summary, simple_diff, summarize_diff
from ldb.path import WorkspacePath
from ldb.status import status
from ldb.utils import WORKSPACE_DATASET_PREFIX, format_dataset_identifier
from ldb.workspace import load_workspace_dataset

if TYPE_CHECKING:
    from argparse import _SubParsersAction


def status_command(options: Namespace) -> None:
    dataset = options.dataset
    ldb_dir = get_ldb_instance()
    ws_status = status(ldb_dir, dataset)
    ds_identifier = format_dataset_identifier(
        ws_status.dataset_name,
        ws_status.dataset_version,
    )
    workspace_path = Path(".")
    prefix = ""
    if not dataset:
        prefix = f"Workspace directory: {os.fspath(workspace_path)!r}\n\n"
    print(
        f"{prefix}"
        f"{ds_identifier}\n"
        f"  Num data objects: {ws_status.num_data_objects:8d}\n"
        f"  Num annotations:  {ws_status.num_annotations:8d}",
    )
    if not dataset:
        workspace_ds = load_workspace_dataset(workspace_path)
        if workspace_ds.parent:
            summary_items = summarize_diff(
                simple_diff(
                    ldb_dir,
                    f"{WORKSPACE_DATASET_PREFIX}{os.fspath(workspace_path)}",
                    workspace_ds.parent,
                ),
            )
        else:
            collection_size = get_collection_size(
                workspace_path / WorkspacePath.COLLECTION,
            )
            summary_items = collection_size, 0, 0
        print()
        if any(summary_items):
            print("Unsaved changes:")
            print(format_summary(*summary_items))
        else:
            print("No unsaved changes.")


def add_parser(
    subparsers: "_SubParsersAction[ArgumentParser]",
    parents: Iterable[ArgumentParser],
) -> None:
    parser = subparsers.add_parser(
        "status",
        parents=parents,
        help="Get status of workspace or dataset",
    )
    parser.add_argument(
        "dataset",
        metavar="<dataset>",
        nargs="?",
        default="",
        help="Dataset to show status for",
    )
    parser.set_defaults(func=status_command)



from argparse import ArgumentParser, Namespace
from typing import TYPE_CHECKING, Iterable

from ldb.core import get_ldb_instance
from ldb.diff import (
    DiffItem,
    DiffType,
    format_summary,
    full_diff,
    simple_diff,
    summarize_diff,
)
from ldb.string_utils import left_truncate
from ldb.utils import DATA_OBJ_ID_PREFIX

if TYPE_CHECKING:
    from argparse import _SubParsersAction


def diff_command(options: Namespace) -> None:
    ldb_dir = get_ldb_instance()
    items = list(
        simple_diff(
            ldb_dir,
            options.dataset1,
            options.dataset2,
            ".",
        ),
    )
    if not options.summary:
        for diff_item in full_diff(ldb_dir, items):
            row = format_diff_item(diff_item, options.verbose)
            if row:
                print(row)
    summary_items = summarize_diff(items)
    if any(summary_items):
        if not options.summary:
            print()
        print(format_summary(*summary_items))


def format_diff_item(diff_item: DiffItem, verbose: bool) -> str:
    if diff_item.diff_type == DiffType.ADDITION:
        prefix = "+"
        annotation_col = annotation_version_str(diff_item.annotation_version2)
    elif diff_item.diff_type == DiffType.DELETION:
        prefix = "-"
        annotation_col = annotation_version_str(diff_item.annotation_version1)
    elif diff_item.diff_type == DiffType.MODIFICATION:
        prefix = "m"
        annotation_col = (
            f"{annotation_version_str(diff_item.annotation_version1)} "
            f"-> {annotation_version_str(diff_item.annotation_version2)}"
        )
    else:
        return ""
    data_object_path = (
        diff_item.data_object_path
        if verbose
        else left_truncate(diff_item.data_object_path)
    )
    return (
        f"{prefix} {DATA_OBJ_ID_PREFIX}{diff_item.data_object_hash:35} "
        f"{annotation_col:12} {data_object_path}"
    )


def annotation_version_str(annotation_version: int) -> str:
    return str(annotation_version) if annotation_version else "-"


def add_parser(
    subparsers: "_SubParsersAction[ArgumentParser]",
    parents: Iterable[ArgumentParser],
) -> None:
    parser = subparsers.add_parser(
        "diff",
        parents=parents,
        help="Show diff of two datasets",
    )
    parser.add_argument(
        "dataset1",
        metavar="<dataset>",
        nargs="?",
        default="",
        help="Dataset to start with",
    )
    parser.add_argument(
        "dataset2",
        metavar="<dataset>",
        nargs="?",
        default="",
        help="Dataset to show changes for",
    )
    parser.add_argument(
        "-s",
        "--summary",
        action="store_true",
        default=False,
        help="Show only the number of additions, deletions and modifications.",
    )
    parser.set_defaults(func=diff_command)



import json
import sys

from .clip_utils import text_similarity
from .utils import sort_by_iterable


def main() -> None:
    if len(sys.argv) < 2:
        print(
            "usage: clip-text <text> [<model_name>]",
            file=sys.stderr,
        )
        sys.exit(1)
    text = sys.argv[1]
    model_name = sys.argv[2] if len(sys.argv) > 2 else None
    data = json.loads(sys.stdin.read())
    if not data:
        return

    file_paths = [d[1] for d in data]
    similarity = text_similarity(text, file_paths, model_name)
    for data_object_hash, _, _ in sort_by_iterable(data, similarity):
        print(data_object_hash, flush=True)


if __name__ == "__main__":
    main()



from argparse import ArgumentParser, Namespace
from typing import TYPE_CHECKING, Iterable

from ldb.add import select_data_object_hashes
from ldb.cli_utils import add_data_obj_params, tag_list
from ldb.core import get_ldb_instance
from ldb.data_object_tag import tag_data_objects
from ldb.exceptions import LDBException
from ldb.utils import DATASET_PREFIX, ROOT

if TYPE_CHECKING:
    from argparse import _SubParsersAction


def tag_command(options: Namespace) -> None:
    paths = options.paths
    query_args = options.query_args
    add_tags = [t for lst in options.add for t in lst]
    remove_tags = [t for lst in options.remove for t in lst]
    if not paths:
        if not query_args:
            raise LDBException(
                "Must provide either a query or at least one path",
            )
        paths = [f"{DATASET_PREFIX}{ROOT}"]
    ldb_dir = get_ldb_instance()
    data_object_hashes = select_data_object_hashes(
        ldb_dir,
        paths,
        query_args,
        warn=False,
    )
    tag_data_objects(
        ldb_dir,
        data_object_hashes,
        add_tags=add_tags,
        remove_tags=remove_tags,
    )


def add_parser(
    subparsers: "_SubParsersAction[ArgumentParser]",
    parents: Iterable[ArgumentParser],
) -> None:
    parser = subparsers.add_parser(
        "tag",
        parents=parents,
        help="Tag data objects",
    )
    add_data_obj_params(parser, dest="query_args")
    parser.add_argument(
        "-a",
        "--add",
        metavar="<tags>",
        default=[],
        type=tag_list,
        action="append",
        help="Comma-separated list of tags to add to data objects",
    )
    parser.add_argument(
        "-r",
        "--remove",
        metavar="<tags>",
        default=[],
        type=tag_list,
        action="append",
        help="Comma-separated list of tags to remove from data objects",
    )
    parser.set_defaults(func=tag_command)



import argparse
from argparse import ArgumentParser

from ldb import __version__
from ldb.command import (
    add,
    add_storage,
    commit,
    completion,
    delete,
    diff,
    ds,
    evaluate,
    index,
    init,
    instantiate,
    ls,
    pull,
    stage,
    status,
    sync,
    tag,
)


def get_main_parser() -> ArgumentParser:
    parent_parser = argparse.ArgumentParser(add_help=False)
    verbosity_group = parent_parser.add_mutually_exclusive_group()
    verbosity_group.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        default=False,
        help="Quiet mode",
    )
    verbosity_group.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Increase verbosity",
    )
    parents = [parent_parser]

    main_parser = argparse.ArgumentParser(
        prog="ldb",
        description="Label Database",
        parents=parents,
    )
    main_parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=f"%(prog)s {__version__}",
    )
    subparsers = main_parser.add_subparsers()
    add.add_parser(subparsers, parents)
    add_storage.add_parser(subparsers, parents)
    completion.add_parser(subparsers, parents)
    commit.add_parser(subparsers, parents)
    delete.add_parser(subparsers, parents)
    diff.add_parser(subparsers, parents)
    ds.add_parser(subparsers, parents)
    evaluate.add_parser(subparsers, parents)
    index.add_parser(subparsers, parents)
    init.add_parser(subparsers, parents)
    instantiate.add_parser(subparsers, parents)
    ls.add_parser(subparsers, parents)
    pull.add_parser(subparsers, parents)
    stage.add_parser(subparsers, parents)
    status.add_parser(subparsers, parents)
    sync.add_parser(subparsers, parents)
    tag.add_parser(subparsers, parents)
    return main_parser



class Format:
    AUTO = "auto-detect"
    STRICT = "strict-pairs"
    BARE = "bare-pairs"
    ANNOT = "annotation-only"
    INFER = "tensorflow-inferred"
    LABEL_STUDIO = "label-studio"


INSTANTIATE_FORMATS = {
    "strict": Format.STRICT,
    Format.STRICT: Format.STRICT,
    "bare": Format.BARE,
    Format.BARE: Format.BARE,
    "infer": Format.INFER,
    Format.INFER: Format.INFER,
    "annot": Format.ANNOT,
    Format.ANNOT: Format.ANNOT,
    Format.LABEL_STUDIO: Format.LABEL_STUDIO,
}
INDEX_FORMATS = {
    "auto": Format.AUTO,
    Format.AUTO: Format.AUTO,
    **INSTANTIATE_FORMATS,
}



import json
from pathlib import Path
from typing import Collection, List, Sequence, Set, Tuple

import jmespath
from fsspec.utils import get_protocol
from funcy.objects import cached_property

from ldb.exceptions import IndexingException
from ldb.index.base import PairIndexer, Preprocessor
from ldb.index.inferred import InferredIndexingItem
from ldb.index.utils import (
    FileSystemPath,
    FSPathsMapping,
    IndexingJobMapping,
    get_annotation_content,
)
from ldb.storage import StorageLocation, get_filesystem
from ldb.typing import JSONObject
from ldb.utils import current_time, get_file_hash


class LabelStudioPreprocessor(Preprocessor):
    def __init__(
        self,
        paths: Sequence[str],
        storage_locations: Sequence[StorageLocation],
        path_key: str = "",
    ) -> None:
        super().__init__(paths, storage_locations)
        self.path_key = path_key

    @cached_property
    def data_object_path_and_annotation_pairs(
        self,
    ) -> List[Tuple[FileSystemPath, JSONObject]]:
        result = []
        for annot_fs, annot_paths in self.annotation_paths.items():
            for annot_path in annot_paths:
                orig_annot = get_annotation_content(annot_fs, annot_path)
                annotations: List[JSONObject]
                if isinstance(orig_annot, list):
                    annotations = orig_annot
                elif isinstance(orig_annot, dict):
                    annotations = [orig_annot]
                else:
                    raise IndexingException(
                        "Annotation file must contain a JSON array or JSON "
                        "object for label-studio format. Incorrectly "
                        f"formatted file: {annot_path}",
                    )
                for annot in annotations:
                    data = annot["data"]
                    data_object_info = data.setdefault("data-object-info", {})

                    path_key = data.get("path_key")
                    if path_key is None:
                        if self.path_key:
                            path_key = self.path_key
                        else:
                            path_key = infer_data_object_path_key(annot)
                            self.path_key = path_key
                            print(f"Inferred data object path key: {path_key}")
                        data_object_info["path_key"] = path_key
                    try:
                        orig_path = jmespath.search(path_key, annot)
                        assert isinstance(orig_path, str)
                    except Exception as exc:
                        raise IndexingException(
                            f"Invalid data object path key: {path_key}",
                        ) from exc

                    protocol = get_protocol(orig_path)
                    obj_fs = get_filesystem(
                        orig_path,
                        protocol,
                        self.storage_locations,
                    )
                    path: str = obj_fs._strip_protocol(  # pylint: disable=protected-access # noqa: E501
                        orig_path,
                    )
                    if "md5" not in data_object_info:
                        data_object_info["md5"] = get_file_hash(obj_fs, path)
                    result.append((FileSystemPath(obj_fs, path), annot))
        return result

    @cached_property
    def annotations(self) -> List[JSONObject]:
        return [a for _, a in self.data_object_path_and_annotation_pairs]

    @cached_property
    def data_object_paths(self) -> FSPathsMapping:
        paths: FSPathsMapping = {}
        for (fs, path), _ in self.data_object_path_and_annotation_pairs:
            paths.setdefault(fs, []).append(path)
        return paths


class LabelStudioIndexer(PairIndexer):
    def __init__(
        self,
        ldb_dir: Path,
        preprocessor: LabelStudioPreprocessor,
        read_any_cloud_location: bool,
        strict_format: bool,
        tags: Collection[str] = (),
    ) -> None:
        self.preprocessor: LabelStudioPreprocessor
        super().__init__(
            ldb_dir,
            preprocessor,
            read_any_cloud_location,
            strict_format,
            tags,
        )

    def _index(self) -> None:
        indexing_jobs, _ = self.process_files()
        self.index_label_studio_files(
            indexing_jobs,
            self.preprocessor.annotations,
            self.tags,
        )

    def index_label_studio_files(
        self,
        indexing_jobs: IndexingJobMapping,
        annotations: List[JSONObject],
        tags: Collection[str],
    ) -> None:
        annot_iter = iter(annotations)

        for fs, jobs in indexing_jobs.items():
            for config, path_seq in jobs:
                for data_object_path in path_seq:
                    obj_result = InferredIndexingItem(
                        self.ldb_dir,
                        current_time(),
                        tags,
                        FileSystemPath(fs, data_object_path),
                        config.save_data_object_path_info,
                        self.hashes,
                        next(annot_iter),
                    ).index_data()
                    self.result.append(obj_result)


def infer_data_object_path_key(annot: JSONObject) -> str:
    try:
        data = annot["data"]
    except KeyError as exc:
        raise IndexingException(
            "Cannot infer data object path key. "
            'Missing top-level "data" key',
        ) from exc
    data_keys: Set[str] = data.keys() - {"data-object-info"}
    if len(data_keys) != 1:
        raise IndexingException(
            "Cannot infer data object path key. "
            "Found multiple possible keys under top-level "
            f'"data" key: {data_keys}',
        )
    subkey = json.dumps(data_keys.pop())
    return f'"data".{subkey}'



from jmespath.parser import ParsedResult

from ldb.jmespath.parser import CustomParser, FieldValidatedTreeInterpreter


def jp_compile(expression: str) -> ParsedResult:
    return CustomParser().parse(expression)


def custom_compile(expression: str) -> ParsedResult:
    return CustomParser(
        tree_interpreter_class=FieldValidatedTreeInterpreter,
    ).parse(expression)



from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import TYPE_CHECKING, Iterable

from ldb.cli_utils import add_data_format_arguments, add_target_dir_argument
from ldb.core import get_ldb_instance
from ldb.data_formats import INSTANTIATE_FORMATS, Format
from ldb.instantiate import instantiate
from ldb.utils import format_dataset_identifier
from ldb.workspace import load_workspace_dataset

if TYPE_CHECKING:
    from argparse import _SubParsersAction


def instantiate_command(options: Namespace) -> None:
    ldb_dir = get_ldb_instance()
    workspace_path = Path(".")
    workspace_ds = load_workspace_dataset(workspace_path)
    ds_ident = format_dataset_identifier(workspace_ds.dataset_name)
    print(f"Instantiating working dataset {ds_ident}...\n")
    result = instantiate(
        ldb_dir,
        workspace_path,
        Path(options.target_dir),
        fmt=options.format,
        force=options.force,
        apply=options.apply,
    )
    print(
        "Copied dataset to workspace.\n"
        f"  Data objects: {result.num_data_objects:9d}\n"
        f"  Annotations:  {result.num_annotations:9d}",
    )


def add_parser(
    subparsers: "_SubParsersAction[ArgumentParser]",
    parents: Iterable[ArgumentParser],
) -> None:
    parser = subparsers.add_parser(
        "instantiate",
        parents=parents,
        help="Instantiate the current workspace dataset",
    )
    add_data_format_arguments(
        parser,
        default=Format.BARE,
        formats=INSTANTIATE_FORMATS,
    )
    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        default=False,
        help="Remove existing workspace contents",
    )
    add_target_dir_argument(parser)
    parser.add_argument(
        "--apply",
        nargs="+",
        metavar="<exec>",
        default=None,
        dest="apply",
        help="Executable to apply to data objects and annotations",
    )
    parser.set_defaults(func=instantiate_command)



from argparse import ArgumentParser, Namespace
from typing import TYPE_CHECKING, Iterable

import shtab

if TYPE_CHECKING:
    from argparse import _SubParsersAction


def completion_command(options: Namespace) -> None:
    from ldb.cli import (  # pylint: disable=import-outside-toplevel,cyclic-import # noqa: E501
        get_main_parser,
    )

    parser = get_main_parser()
    script = shtab.complete(parser, shell=options.shell)
    print(script)


def add_parser(
    subparsers: "_SubParsersAction[ArgumentParser]",
    parents: Iterable[ArgumentParser],
) -> None:
    parser = subparsers.add_parser(
        "completion",
        parents=parents,
        help="Output shell completion script",
    )
    parser.add_argument(
        "-s",
        "--shell",
        help="Shell syntax for completions.",
        default="bash",
        choices=shtab.SUPPORTED_SHELLS,
    )
    parser.set_defaults(func=completion_command)



import os
from typing import Any, Dict, Optional, Sequence, Union

from fsspec.callbacks import _DEFAULT_CALLBACK, Callback
from fsspec.spec import AbstractFileSystem

FSProtocol = Union[str, Sequence[str]]


def first_protocol(fs_protocol: FSProtocol) -> str:
    if isinstance(fs_protocol, str):
        return fs_protocol
    return fs_protocol[0]


def has_protocol(fs_protocol: FSProtocol, protocol: str) -> bool:
    if isinstance(fs_protocol, str):
        return protocol == fs_protocol
    return protocol in fs_protocol


def cp_file_any_fs(
    source_fs: AbstractFileSystem,
    source_path: str,
    dest_fs: AbstractFileSystem,
    dest_path: str,
) -> None:
    if source_fs is dest_fs:
        source_fs.cp_file(source_path, dest_path)
    elif source_fs.protocol == "file":
        dest_fs.put_file(source_path, dest_path)
    elif dest_fs.protocol == "file":
        source_fs.get_file(source_path, dest_path)
    else:
        cp_file_across_fs(source_fs, source_path, dest_fs, dest_path)


def cp_file_across_fs(
    source_fs: AbstractFileSystem,
    source_path: str,
    dest_fs: AbstractFileSystem,
    dest_path: str,
    callback: Callback = _DEFAULT_CALLBACK,
    source_kwargs: Optional[Dict[str, Any]] = None,
    dest_kwargs: Optional[Dict[str, Any]] = None,
) -> None:
    """
    Transfer single file between any two filesystems.

    This implementation combines code from:
    fsspec.spec.AbstractFileSystem.put_file
    fsspec.spec.AbstractFileSystem.get_file
    """
    if source_fs.isdir(source_path):
        dest_fs.makedirs(dest_path, exist_ok=True)
        return None

    if source_kwargs is None:
        source_kwargs = {}
    if dest_kwargs is None:
        dest_kwargs = {}
    with source_fs.open(source_path, "rb", **source_kwargs) as f1:
        callback.set_size(getattr(f1, "size", None))
        parent = dest_fs._parent(  # pylint: disable=protected-access
            os.fspath(dest_path),
        )
        dest_fs.mkdirs(parent, exist_ok=True)
        with dest_fs.open(dest_path, "wb", **dest_kwargs) as f2:
            data = True
            while data:
                data = f1.read(source_fs.blocksize)
                segment_len = f2.write(data)
                callback.relative_update(segment_len)



from dataclasses import dataclass

from funcy.objects import cached_property

from ldb.exceptions import DataObjectNotFoundError, IndexingException
from ldb.index.base import (
    AnnotationFileIndexingItem,
    IndexedObjectResult,
    Indexer,
)
from ldb.index.utils import (
    DataObjectMeta,
    FileSystemPath,
    get_annotation_content,
)
from ldb.typing import JSONObject
from ldb.utils import DATA_OBJ_ID_PREFIX, current_time, load_data_file


class AnnotationOnlyIndexer(Indexer):
    def _index(self) -> None:
        for fs, paths in self.preprocessor.annotation_paths.items():
            for path in paths:
                item = AnnotationOnlyIndexingItem(
                    self.ldb_dir,
                    current_time(),
                    self.tags,
                    FileSystemPath(fs, path),
                )
                self.result.append(item.index_data())


@dataclass
class AnnotationOnlyIndexingItem(AnnotationFileIndexingItem):
    annotation_fsp: FileSystemPath

    @cached_property
    def annotation_file_content(self) -> JSONObject:
        return get_annotation_content(  # type: ignore[return-value]
            *self.annotation_fsp,
        )

    @cached_property
    def data_object_hash(self) -> str:
        try:
            return self.annotation_file_content[  # type: ignore[no-any-return]
                "ldb_meta"
            ]["data_object_id"]
        except KeyError as exc:
            raise IndexingException(
                "Missing ldb_meta.data_object_id key for annotation-only "
                f"format: {self.annotation_fsp.path}",
            ) from exc

    @cached_property
    def data_object_meta(self) -> DataObjectMeta:
        meta_content: DataObjectMeta = load_data_file(
            self.data_object_meta_file_path,
        )
        meta_content["last_indexed"] = self.current_timestamp
        meta_content["tags"] = sorted(  # type: ignore[assignment]
            set(meta_content["tags"])  # type: ignore[arg-type]
            | set(self.tags),
        )
        return meta_content

    @cached_property
    def annotation_content(self) -> JSONObject:  # type: ignore[override]
        return self.annotation_file_content["annotation"]  # type: ignore[no-any-return] # noqa: E501

    def index_data(self) -> IndexedObjectResult:
        if not self.data_object_dir.exists():
            raise DataObjectNotFoundError(
                "Data object not found: "
                f"{DATA_OBJ_ID_PREFIX}{self.data_object_hash} "
                f"(annotation_file_path={self.annotation_fsp.path!r})",
            )

        new_annotation = not self.annotation_meta_file_path.is_file()
        self.enqueue_data(self.data_object_to_write())
        self.enqueue_data(self.annotation_to_write())
        self.write_data()
        return IndexedObjectResult(
            found_data_object=False,
            found_annotation=True,
            new_data_object=False,
            new_annotation=new_annotation,
            data_object_hash=self.data_object_hash,
        )



import os
from contextlib import contextmanager
from pathlib import Path
from typing import Callable, Dict, Generator, List, Optional, Sequence

from appdirs import site_config_dir, user_config_dir
from tomlkit import document, dumps, parse
from tomlkit.exceptions import NonExistentKey
from tomlkit.toml_document import TOMLDocument

from ldb.app_info import APP_AUTHOR, APP_NAME
from ldb.env import Env
from ldb.exceptions import LDBException, LDBInstanceNotFoundError
from ldb.path import DirName, Filename, GlobalDir


class ConfigType:
    INSTANCE = "INSTANCE"
    USER = "USER"
    SYSTEM = "SYSTEM"


DEFAULT_CONFIG_TYPES = (
    ConfigType.INSTANCE,
    ConfigType.USER,
    ConfigType.SYSTEM,
)
GLOBAL_CONFIG_TYPES = (
    ConfigType.USER,
    ConfigType.SYSTEM,
)


def load_from_path(path: Path) -> TOMLDocument:
    with path.open() as file:
        config_str = file.read()
    return parse(config_str)


def save_to_path(config: TOMLDocument, path: Path) -> None:
    config_str = dumps(config)
    if not path.parent.is_dir():
        path.parent.mkdir(parents=True, exist_ok=True)
    with path.open("w") as file:
        file.write(config_str)


def load_first(
    config_types: Sequence[str] = DEFAULT_CONFIG_TYPES,
) -> Optional[TOMLDocument]:
    for config_type in config_types:
        for config_dir in get_config_dirs(config_type):
            path = config_dir / Filename.CONFIG
            try:
                return load_from_path(path)
            except FileNotFoundError:
                pass
    return None


@contextmanager
def edit(path: Path) -> Generator[TOMLDocument, None, None]:
    try:
        config = load_from_path(path)
    except FileNotFoundError:
        config = document()
    yield config
    save_to_path(config, path)


def _get_global_base_parent() -> Path:
    return Path.home()


def get_global_base() -> Path:
    return _get_global_base_parent() / DirName.GLOBAL_BASE


def get_default_instance_dir() -> Path:
    return get_global_base() / GlobalDir.DEFAULT_INSTANCE


def get_ldb_dir() -> Path:
    """
    Get the directory we should use as the ldb instance.
    """
    if Env.LDB_DIR in os.environ:
        return Path(os.environ[Env.LDB_DIR])
    config = load_first(GLOBAL_CONFIG_TYPES)
    if config is not None:
        try:
            ldb_dir_str: str = config["core"]["ldb_dir"]  # type: ignore[index,assignment] # noqa: E501
        except NonExistentKey:
            pass
        else:
            ldb_dir_path = Path(ldb_dir_str).expanduser()
            if ldb_dir_path.is_absolute():
                return ldb_dir_path
            raise LDBException(
                "Found relative path for core.ldb_dir: {repr(ldb_dir_str)}"
                "Paths in LDB config must be absolute",
            )
    return get_default_instance_dir()


def get_instance_config_dirs() -> List[Path]:
    try:
        return [get_ldb_dir()]
    except LDBInstanceNotFoundError:
        return []


def get_user_config_dirs() -> List[Path]:
    config_dirs = [get_global_base()]
    try:
        additional_dir = user_config_dir(APP_NAME, APP_AUTHOR)
    except OSError:
        pass
    else:
        config_dirs.append(Path(additional_dir))
    return config_dirs


def get_system_config_dirs() -> List[Path]:
    config_dirs = []
    try:
        additional_dir = site_config_dir(APP_NAME, APP_AUTHOR)
    except OSError:
        pass
    else:
        config_dirs.append(Path(additional_dir))
    return config_dirs


CONFIG_DIR_FUNCTIONS: Dict[str, Callable[[], List[Path]]] = {
    ConfigType.INSTANCE: get_instance_config_dirs,
    ConfigType.USER: get_user_config_dirs,
    ConfigType.SYSTEM: get_system_config_dirs,
}


def get_config_dirs(config_type: str) -> List[Path]:
    return CONFIG_DIR_FUNCTIONS[config_type]()


def set_default_instance(path: Path, overwrite_existing: bool = False) -> None:
    path = path.absolute()
    config_dir = get_global_base()
    config_dir.mkdir(parents=True, exist_ok=True)
    config_file = config_dir / Filename.CONFIG
    with edit(config_file) as cfg:
        if "core" in cfg:
            if (
                not overwrite_existing
                and "ldb_dir" in cfg["core"]  # type: ignore[operator]
            ):
                value: str = cfg["core"][  # type: ignore[index,assignment]
                    "ldb_dir"
                ]
                print(
                    "Not setting core.ldb_dir as it is already set "
                    f"to {repr(value)}",
                )
                return
        else:
            cfg["core"] = {}
        new_value = os.fspath(path)
        cfg["core"]["ldb_dir"] = new_value  # type: ignore[index]
    print(f"Set core.ldb_dir to {repr(new_value)}")



from functools import partial
from itertools import tee
from pathlib import Path
from typing import Iterable, Iterator, List, Optional, Sequence, Tuple, Union

from ldb.add import process_args_for_ls
from ldb.dataset import apply_queries, get_annotations, get_data_object_metas
from ldb.func_utils import apply_optional
from ldb.op_type import OpType
from ldb.query.search import SearchFunc, get_search_func
from ldb.typing import JSONDecoded

EvaluateResult = Union[
    Tuple[str, JSONDecoded],
    Tuple[str, JSONDecoded, JSONDecoded],
]
QuerySubject = Union[List[str], List[Optional[str]]]


def evaluate(
    ldb_dir: Path,
    paths: Sequence[str],
    collection_ops: Iterable[Tuple[str, str]],
    warn: bool = True,
) -> Iterator[EvaluateResult]:
    search, file_search, collection_ops = process_query_args(
        collection_ops,
    )
    (
        data_object_hashes,
        annotation_hashes,
        _,
    ) = process_args_for_ls(  # pylint: disable=duplicate-code
        ldb_dir,
        paths,
    )
    collection = apply_queries(
        ldb_dir,
        data_object_hashes,
        annotation_hashes,
        collection_ops,
        warn=warn,
    )
    collection1, collection2 = tee(collection)
    data_object_hashes1, data_object_hashes2 = tee(d for d, _ in collection1)
    annotation_hashes = (a for _, a in collection2)
    if search is None and file_search is None:
        search_results = [get_annotations(ldb_dir, annotation_hashes)]
    else:
        search_results = []
        if file_search is not None:
            search_results.append(
                file_search(  # type: ignore[arg-type]
                    get_data_object_metas(ldb_dir, data_object_hashes1),
                ),
            )
        if search is not None:
            search_results.append(
                search(get_annotations(ldb_dir, annotation_hashes)),  # type: ignore[arg-type] # noqa: E501
            )

    result: Iterator[EvaluateResult] = zip(  # type: ignore[assignment]
        data_object_hashes2,
        *search_results,
    )
    yield from result


def process_query_args(
    query_args: Iterable[Tuple[str, str]],
    warn: bool = True,
) -> Tuple[Optional[SearchFunc], Optional[SearchFunc], List[Tuple[str, str]]]:
    query_args = list(query_args)
    search = None
    file_search = None
    if not query_args:
        return search, file_search, query_args
    op_type, arg = query_args.pop()
    if op_type == OpType.ANNOTATION_QUERY:
        search = apply_optional(
            partial(get_search_func, use_custom=True, warn=warn),
            arg,
        )
    elif op_type == OpType.JP_ANNOTATION_QUERY:
        search = apply_optional(
            partial(get_search_func, use_custom=False, warn=warn),
            arg,
        )
    elif op_type == OpType.FILE_QUERY:
        file_search = apply_optional(
            partial(get_search_func, warn=warn),
            arg,
        )
    else:
        query_args.append((op_type, arg))
    return search, file_search, query_args



import shlex
from typing import NoReturn


def dependency_exc(exc: Exception, extra: str) -> NoReturn:
    package = shlex.quote(f"ldb-alpha[{extra}]")
    raise ImportError(
        "Missing dependency. To install all dependencies run:\n\n"
        f"\tpip install {package}\n",
    ) from exc


def validate_clip() -> None:
    try:
        import clip  # pylint: disable=import-outside-toplevel,unused-import # noqa: F401,E501
    except ModuleNotFoundError as exc:
        dependency_exc(exc, "clip-plugin")


def validate_resnet() -> None:
    try:
        import torch  # pylint: disable=import-outside-toplevel,unused-import # noqa: F401,E501
        import torchvision  # pylint: disable=import-outside-toplevel,unused-import # noqa: F401,E501
    except ModuleNotFoundError as exc:
        dependency_exc(exc, "clip-plugin")


def clip_text() -> None:
    validate_clip()

    from .clip_text import main  # pylint: disable=import-outside-toplevel

    main()


def clip_image() -> None:
    validate_clip()

    from .clip_image import main  # pylint: disable=import-outside-toplevel

    main()


def resnet_image() -> None:
    validate_resnet()

    from .resnet_image import main  # pylint: disable=import-outside-toplevel

    main()



from dataclasses import dataclass
from itertools import chain
from pathlib import Path
from typing import Collection, Dict, List, Tuple, Union

from fsspec.spec import AbstractFileSystem
from funcy.objects import cached_property

from ldb.data_formats import Format
from ldb.exceptions import IndexingException
from ldb.index.base import (
    DataObjectFileIndexingItem,
    PairIndexer,
    Preprocessor,
)
from ldb.index.utils import (
    AnnotationMeta,
    FileSystemPath,
    FSPathsMapping,
    IndexingJobMapping,
    construct_annotation_meta,
    expand_dir_paths,
    is_hidden_fsspec_path,
)
from ldb.typing import JSONDecoded, JSONObject
from ldb.utils import current_time, load_data_file


class InferredPreprocessor(Preprocessor):
    @cached_property
    def dir_path_to_files(
        self,
    ) -> Dict[AbstractFileSystem, Dict[str, List[str]]]:
        dir_path_lists = expand_dir_paths(self.paths)
        file_seqs: Dict[AbstractFileSystem, Dict[str, List[str]]] = {}
        for fs, dir_paths in dir_path_lists.items():
            fs_seq: Dict[str, List[str]] = file_seqs.setdefault(fs, {})
            for p in dir_paths:
                file_paths: List[str] = [
                    p
                    for p in fs.find(p.rstrip("/"))
                    if not is_hidden_fsspec_path(p)
                ]
                if file_paths:
                    fs_seq[p] = file_paths
        return file_seqs

    def get_storage_files(self) -> Dict[AbstractFileSystem, List[str]]:
        return {
            fs: list(chain(*seqs.values()))
            for fs, seqs in self.dir_path_to_files.items()
        }

    @cached_property
    def files_by_type(self) -> Tuple[FSPathsMapping, FSPathsMapping]:
        files, annotation_files = super().files_by_type
        for fs, path_seq in annotation_files.items():
            if path_seq:
                first_path = path_seq[0]
                raise IndexingException(
                    "No annotation files should be present for "
                    f"{Format.INFER} format.\n"
                    f"Found {len(annotation_files)} JSON files.\n"
                    f"First path: {first_path} (protocol={fs.protocol})",
                )
        return files, annotation_files

    @cached_property
    def data_object_paths(self) -> FSPathsMapping:
        return {
            fs: [
                path for dir_paths in path_maps.values() for path in dir_paths
            ]
            for fs, path_maps in self.dir_path_to_files.items()
        }


class InferredIndexer(PairIndexer):
    def __init__(
        self,
        ldb_dir: Path,
        preprocessor: InferredPreprocessor,
        read_any_cloud_location: bool,
        strict_format: bool,
        tags: Collection[str] = (),
    ) -> None:
        self.preprocessor: InferredPreprocessor
        super().__init__(
            ldb_dir,
            preprocessor,
            read_any_cloud_location,
            strict_format,
            tags,
        )

    def infer_annotations(
        self,
    ) -> Dict[AbstractFileSystem, Dict[str, JSONObject]]:
        annotations: Dict[AbstractFileSystem, Dict[str, JSONObject]] = {}
        for fs, seqs in self.preprocessor.dir_path_to_files.items():
            fs_annots: Dict[str, JSONObject] = annotations.setdefault(fs, {})
            for dir_path, file_seq in seqs.items():
                len_dir_path = len(dir_path)
                for file in file_seq:
                    raw_label = (
                        file[len_dir_path:].rsplit("/", 1)[0].strip("/")
                    )
                    label_parts = raw_label.lstrip("/").split("/")
                    label: Union[str, JSONObject] = label_parts[-1]
                    for p in label_parts[-2::-1]:
                        label = {p: label}
                    fs_annots[file] = {"label": label}
        return annotations

    def _index(self) -> None:
        annotations = self.infer_annotations()
        indexing_jobs, _ = self.process_files()

        annotations_by_data_object_path: Dict[
            Tuple[AbstractFileSystem, str],
            JSONObject,
        ] = {}

        for fs, fs_annotations in annotations.items():
            fs_old_to_new = self.old_to_new_files.get(fs, {})
            for f, annot in fs_annotations.items():
                path = fs_old_to_new.get(f)
                if path is None:
                    path = FileSystemPath(fs, f)
                annotations_by_data_object_path[path] = annot
        self.index_inferred_files(
            indexing_jobs,
            annotations_by_data_object_path,
            self.tags,
        )

    def index_inferred_files(
        self,
        indexing_jobs: IndexingJobMapping,
        annotations_by_data_object_path: Dict[
            Tuple[AbstractFileSystem, str],
            JSONObject,
        ],
        tags: Collection[str],
    ) -> None:
        for fs, jobs in indexing_jobs.items():
            for config, path_seq in jobs:
                for data_object_path in path_seq:
                    item = InferredIndexingItem(
                        self.ldb_dir,
                        current_time(),
                        tags,
                        FileSystemPath(fs, data_object_path),
                        config.save_data_object_path_info,
                        self.hashes,
                        annotations_by_data_object_path[fs, data_object_path],
                    )
                    self.result.append(item.index_data())


@dataclass
class InferredIndexingItem(DataObjectFileIndexingItem):
    _annotation_content: JSONDecoded

    @cached_property
    def annotation_meta(self) -> AnnotationMeta:
        prev_annotation = (
            load_data_file(self.annotation_meta_file_path)
            if self.annotation_meta_file_path.exists()
            else {}
        )
        return construct_annotation_meta(
            prev_annotation,
            self.current_timestamp,
            self.annotation_version,
            self.curr_time,
        )

    @cached_property
    def annotation_content(self) -> JSONDecoded:
        return self._annotation_content

    @cached_property
    def has_annotation(self) -> bool:
        return True
